package com.jetlore.sdk.common;


import android.os.SystemClock;
import android.util.Log;

import com.jetlore.sdk.http.AsyncHttpClient;
import com.jetlore.sdk.http.AsyncHttpResponseHandler;
import com.jetlore.sdk.http.RequestHandle;
import com.jetlore.sdk.http.ResponseHandlerInterface;

import org.apache.http.Header;
import org.apache.http.HttpResponse;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public abstract class Event {

    public static final String TAG = "JetloreActionTracker";

    protected String        userId;
    protected String        apiUrl;
    protected AsyncHttpClient httpClient;

    private String          feedId;
    private List<String>    labels;
    private Map<String, String> customParameters;

    protected Event(EventBuilder<? extends Event> builder) {
        this.userId = builder.userId;
        this.feedId = builder.feedId;
        this.labels = builder.labels;

        this.apiUrl = builder.apiUrl;
        this.httpClient = builder.httpClient;
        this.customParameters = builder.customParameters;
    }

    protected abstract Map<String, String> getQueryParams();

    protected void onSuccess(int statusCode, Header[] headers, byte[] responseBody) {}
    protected void onFailure(int statusCode, Header[] headers, byte[] responseBody, Throwable error) {}

    protected abstract  String generateMainGetRequestURL();

    protected String toURL() {

        Map<String, String> queryParams = getQueryParams();

        StringBuilder sb = new StringBuilder(generateMainGetRequestURL());
        if (!Util.isNullOrEmpty(feedId)) {
            queryParams.put("feed", feedId);
        }

        if (null != labels && !labels.isEmpty()) {
            Iterator<String> labelsIt= labels.iterator();
            StringBuilder lbs = new StringBuilder();
            lbs.append(labelsIt.next());

            while(labelsIt.hasNext()) {
                lbs.append(",");
                lbs.append(labelsIt.next());
            }
            queryParams.put("jl_labels", lbs.toString());
        }

        if (null != queryParams) {
            for (Map.Entry<String, String> parameter : queryParams.entrySet()) {
                sb.append(String.format("&%s=%s", Util.URLEncode(parameter.getKey()), Util.URLEncode(parameter.getValue())));
            }
        }

        if (null != customParameters) {
            // prevent parameter intersections
            for (String customKey : customParameters.keySet()) {
                if (null != queryParams && queryParams.containsKey(customKey)) {
                    throw new InvalidUseException("Custom parameter name '" + customKey + "' intersects with internally used. " +
                            "Please use different name.");
                }
            }

            for (Map.Entry<String, String> parameter : customParameters.entrySet()) {
                sb.append(String.format("&%s=%s", Util.URLEncode(parameter.getKey()), Util.URLEncode(parameter.getValue())));
            }
        }

        return sb.toString();
    }

    /**
     * Actual call to trigger event. Please see {@link }
     */
    protected abstract void fire();

    public static abstract class EventBuilder<T> {

        private String          userId;
        private String          apiUrl;
        private AsyncHttpClient httpClient;

        private String        feedId;
        private List<String>  labels = new LinkedList<String>();
        private Map<String, String> customParameters = new HashMap<String, String>();

        protected EventBuilder(String userId, String apiUrl, AsyncHttpClient httpClient) {
            this.userId = userId;
            this.apiUrl = apiUrl;
            this.httpClient = httpClient;
        }

        public EventBuilder feed(String feedId){
            this.feedId = feedId;
            return this;
        }

        public EventBuilder label(String label){
            labels.add(label);
            return this;
        }

        public EventBuilder labels(List<String> labels){
            this.labels.addAll(labels);
            return this;
        }

        public EventBuilder customParameter(String name, String value){
            customParameters.put(name, value);
            return this;
        }

        public EventBuilder customParameters(Map<String, String> parameters){
            customParameters.putAll(parameters);
            return this;
        }

        public abstract T build();
    }


    /*
    Request class is a wrapper to allow logarithmic retries in case the server returns 5xx
    maximum number of retries is 5, the delays before requests are 1s, 2s, 4s, 8s, 16s
    in case of other internet access issues, retries are handled by com.jetlore.sdk.http.RetryHandler
     */
    protected abstract class Request {
        RequestHandle _request;
        int _numberOfRetries = 5;
        int _waitMillis = 1000;

        protected Request() { }

        protected abstract RequestHandle send(AsyncHttpResponseHandler responseHandler);
        protected abstract void onSuccess(int statusCode, Header[] headers, byte[] responseBody);
        protected abstract void onFailure(int statusCode, Header[] headers, byte[] responseBody, Throwable error);

        private boolean shouldRetry(HttpResponse response) {
            int statusCode = response.getStatusLine().getStatusCode();
            return _numberOfRetries > 0 && statusCode >= 500 && statusCode <= 599;
        }

        public void run() {
            final Request self = this;
            final AsyncHttpResponseHandler responseHandler = new AsyncHttpResponseHandler() {
                @Override
                public void onPreProcessResponse(ResponseHandlerInterface instance, HttpResponse response) {
                    if (shouldRetry(response)) {
                        Log.d(TAG, String.format("request %s failed, %d retries left, wait %d ms before retry", this.getRequestURI(), _numberOfRetries, _waitMillis));
                        // cancel the current request so that onSuccess / onFailure are not called
                        if (_request != null) { _request.cancel(false); }
                        // onPreProcessResponse is called on AsyncHttpClient thread, so sleep won't block UI
                        // it's OK to block AsyncHttpClient thread since there is no sense in processing
                        // other requests if current failed
                        SystemClock.sleep(_waitMillis);
                        --_numberOfRetries;
                        _waitMillis *= 2;
                        _request = send(this);
                    }
                }

                @Override
                public void onSuccess(int statusCode, Header[] headers, byte[] responseBody) {
                    self.onSuccess(statusCode, headers, responseBody);
                }

                @Override
                public void onFailure(int statusCode, Header[] headers, byte[] responseBody, Throwable error) {
                    self.onFailure(statusCode, headers, responseBody, error);
                }

                @Override
                public void onCancel() {
                    // cancel is used before retrying request to prevent onSuccess / onFailure
                    // onCancel() is empty to avoid unnecessary logging
                }
            };
            _request = send(responseHandler);
        }
    }
}
