package com.jetlore.sdk.common;


import android.content.Context;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.support.annotation.NonNull;
import android.util.Log;

import com.jetlore.sdk.http.AsyncHttpClient;
import com.jetlore.sdk.http.PersistentCookieStore;
import com.jetlore.sdk.tracking.ActionContext;

public abstract class JetloreCommon {

    private static final String TAG = "JetloreCommon";

    /** Constant value to use with {@link #configure} method if user id is unknown at the moment */
    public final static String NO_USER = "no-user";

    protected String _apiUrl;
    protected String _userId;

    protected AsyncHttpClient _httpClient;
    protected Context _context;

    /** Creates a new ActionTracker
     *
     * <p>Uses provided {@link android.content.Context} to get Jetlore API access key
     * from AndroidManifest.xml and to get {@link android.content.SharedPreferences} to persist cookies
     */
    public JetloreCommon(@NonNull Context context) {
        initApiUrlAndAccessTokenFromManifest(context);
        initAsyncHttpClient(context);
    }

    private void initAsyncHttpClient(@NonNull Context context) {
        _context = context;
        _httpClient = new AsyncHttpClient();
        PersistentCookieStore cookieStore = new PersistentCookieStore(context);
        _httpClient.setCookieStore(cookieStore);
    }

    private void initApiUrlAndAccessTokenFromManifest(@NonNull Context context) {
        String apiUrl;

        ApplicationInfo applicationInfo;
        try {
            PackageManager packageManager = context.getPackageManager();
            String packageName = context.getPackageName();
            applicationInfo = packageManager.getApplicationInfo(packageName, PackageManager.GET_META_DATA);
        } catch (PackageManager.NameNotFoundException ex) { // can't happen
            throw new RuntimeException("context.getPackageManager().getApplicationInfo(context.getPackageName(), PackageManager.GET_META_DATA) thrown PackageManager.NameNotFoundException", ex);
        }
        Bundle bundle = applicationInfo.metaData;

        InvalidUseException exception = new InvalidUseException("Provide your Jetlore API URL in the AndroidManifest.xml (<meta-data android:name=\"com.jetlore.sdk.API_URL\" android:value=\"<Jetlore API URL>\"/>");
        if (bundle == null) {
            throw exception;
        }

        apiUrl = bundle.getString("com.jetlore.sdk.API_URL");
        if (apiUrl == null) apiUrl = "https://api.jetlore.com";

        _apiUrl = apiUrl;

        readAdditionalParameters(bundle);
    }

    protected abstract void readAdditionalParameters(Bundle bundle);

    /** Initialize ActionTracker with current user id
     *
     * <p>This method can be safely called multiple times (in case user id has changed)</p>
     *
     * <p>Use {@link #NO_USER} as userId if user id is unknown at the moment.
     * Make sure to call {@link #configure} again once user id is known.</p>
     *
     * @param userId The ID of the user, has to be the same ID you use inside of your email template
     *
     *  @throws InvalidUseException
     *                if userId is null or empty
     */
    public void configure(@NonNull String userId) {
        Log.d(TAG, String.format("configure userId: %s", userId));

        if (Util.isNullOrEmpty(userId)) {
            throw new InvalidUseException("userId is null or empty");
        }

        _userId = userId;
    }

    /** Re-Initialize ActionTracker with apiUrl, accessToken and clientHash
     *
     *  Please use {@link #configure(String)} to change or initialize userId
     *  <p>Use {@link #NO_USER} as userId if user id is unknown at the moment.
     *  Make sure to call {@link #configure} again once user id is known.</p>
     *
     *  @throws InvalidUseException
     *                if userId is null or empty
     *                if apiUrl is null or empty
     *                if accessToken is null or empty
     *                if clientHash is null or empty
     */
    public void configure(@NonNull String userId, @NonNull String apiUrl) {
        Log.d(TAG, String.format("configure userId: %s, apiUrl: %s",  userId, apiUrl));

        if (Util.isNullOrEmpty(userId)) {
            throw new InvalidUseException("userId is null or empty");
        }
        if (Util.isNullOrEmpty(apiUrl)) {
            throw new InvalidUseException("apiUrl is null or empty");
        }

        _userId = userId;
        _apiUrl = apiUrl;
    }

    protected void fillActionContextInfo(@NonNull ActionContext actionContext,  GetEvent.EventBuilder<?> builder) {
        if (null == actionContext) {
            throw new InvalidUseException("actionContext is required");
        }

        builder.labels(actionContext.getLabels())
                .feed(actionContext.getFeedId())
                .customParameters(actionContext.getCustomParameters());
    }
}
