package com.jetlore.sdk.common;

import android.util.Log;

import com.jetlore.sdk.http.AsyncHttpResponseHandler;
import com.jetlore.sdk.http.RequestHandle;

import org.apache.http.Header;
import org.apache.http.entity.ByteArrayEntity;

/**
 * Class that triggers HTTP POST calls on {@link Event#fire()}
 */
public abstract class PostEvent extends Event {

    protected PostEvent(EventBuilder<? extends PostEvent> builder) {
       super(builder);
    }

    /**
     * Generates Request body
     * @return body content of request to pass
     *
     * @throws Exception in case some error during body creation happened
     */
    protected abstract byte[] getPostBody() throws Exception;

    /**
     * HTTP Content-Type value
     * @return body content type. Example "application/json"
     */
    protected abstract String getBodyContentType();


    public void fire() {
        try {
            final String url = toURL();
            final byte[] body = getPostBody();
            final String contentType = getBodyContentType();

            Log.d(TAG, String.format("%s request: POST %s", this.getClass().getSimpleName(), url));

            new Request() {
                @Override
                protected RequestHandle send(AsyncHttpResponseHandler responseHandler) {
                    // Context is null - this is fine we will not use
                    return httpClient.post(null, url, new ByteArrayEntity(body), contentType, responseHandler);
                }

                @Override
                public void onSuccess(int statusCode, Header[] headers, byte[] responseBody) {
                    Log.d(TAG, String.format("POST request to %s finished with statusCode %d", url, statusCode));
                    PostEvent.this.onSuccess(statusCode, headers, responseBody);
                }

                @Override
                public void onFailure(int statusCode, Header[] headers, byte[] responseBody, Throwable error) {
                    Log.d(TAG, String.format("POST request to %s finished with statusCode %d and error %s, Response body: %s",
                            url, statusCode, error, (null != responseBody) ? new String(responseBody) : "N/A"));
                    PostEvent.this.onFailure(statusCode, headers, responseBody, error);
                }
            }.run();
        } catch (Exception ex) {
            Log.e(TAG, "Failed to perform POST call", ex);
            PostEvent.this.onFailure(-1, null, null, ex);
        }
    }
}
