package com.jetlore.sdk.ranking;


import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.math.BigDecimal;
import java.util.LinkedList;
import java.util.List;

/**
 * Generic filter. This object is used to filter catalog.
 *
 * Please use {@link Filter.Builder} to create instance
 */
public class Filter {

    private List<String> includeCategories;
    private List<String> excludeCategories;

    private List<String> l1_categories;
    private List<String> l2_categories;
    private List<String> l3_categories;
    private List<String> l4_categories;

    private BigDecimal minPrice;
    private BigDecimal maxPrice;

    private Integer minDiscount;
    private Integer maxDiscount;

    private List<BooleanField> booleanFields;
    private List<TextField>    textFields;
    private List<ListField>    listFields;
    private List<NumericField> numericFields;


    private Filter(Builder builder) {
        this.includeCategories = builder.includeCategories;
        this.excludeCategories = builder.excludeCategories;

        this.l1_categories = builder.l1_categories;
        this.l2_categories = builder.l2_categories;
        this.l3_categories = builder.l3_categories;
        this.l4_categories = builder.l4_categories;

        this.minPrice = builder.minPrice;
        this.maxPrice = builder.maxPrice;

        this.minDiscount = builder.minDiscount;
        this.maxDiscount = builder.maxDiscount;

        this.booleanFields = builder.booleanFields;
        this.textFields = builder.textFields;
        this.listFields = builder.listFields;
        this.numericFields = builder.numericFields;
    }

    protected JSONObject toJSONObject() throws JSONException {
        JSONObject obj = new JSONObject();

        // general category filter
        if (null != includeCategories || null != excludeCategories) {
            JSONObject catFilter = new JSONObject();
            if (null != includeCategories) {
                catFilter.put("include", new JSONArray(includeCategories));
            }
            if (null != excludeCategories) {
                catFilter.put("exclude", new JSONArray(excludeCategories));
            }
            obj.put("category", catFilter);
        }

        // per category filter
        if (null != l1_categories) {
            obj.put("category1", new JSONObject().put("include", new JSONArray(l1_categories)));
        }
        if (null != l2_categories) {
            obj.put("category2", new JSONObject().put("include", new JSONArray(l2_categories)));
        }
        if (null != l3_categories) {
            obj.put("category3", new JSONObject().put("include", new JSONArray(l3_categories)));
        }
        if (null != l4_categories) {
            obj.put("category4", new JSONObject().put("include", new JSONArray(l4_categories)));
        }

        // price filter
        if (null != minPrice || null != maxPrice) {
            JSONObject priceFilter = new JSONObject();
            if (null != minPrice) {
                priceFilter.put("min", minPrice);
            }
            if (null != maxPrice) {
                priceFilter.put("max", maxPrice);
            }
            obj.put("price", priceFilter);
        }

        // discount filter
        if (null != minDiscount || null != maxDiscount) {
            JSONObject discountFilter = new JSONObject();
            if (null != minDiscount) {
                discountFilter.put("min", minDiscount);
            }
            if (null != maxDiscount) {
                discountFilter.put("max", maxDiscount);
            }
            obj.put("discount", discountFilter);
        }

        for(BooleanField value : booleanFields)  value.saveJSONObject(obj);
        for(TextField value : textFields)  value.saveJSONObject(obj);
        for(ListField value : listFields)  value.saveJSONObject(obj);
        for(NumericField value : numericFields)  value.saveJSONObject(obj);

        return obj;
    }

    /**
     * Boolean filed
     */
    static public class BooleanField {
        private String name;
        private boolean value;

        public BooleanField(String name, boolean value) {
            this.name = name;
            this.value = value;
        }

        protected void saveJSONObject(JSONObject obj) throws JSONException {
            obj.put(name, value);
        }
    }

    /**
     * Text filed
     */
    static public class TextField {
        private String name;
        private List<String> values;
        private boolean isInclude;

        /**
         * Constructs text filter filed
         *
         * @param name filed name
         * @param values list of values
         * @param isInclude true in case if you want to use values as include filter criteria, false if exclude
         */
        public TextField(String name, List<String> values, boolean isInclude) {
            this.values = values;
            this.isInclude = isInclude;
            this.name = name;
        }
        protected void saveJSONObject(JSONObject obj) throws JSONException {
            JSONObject textFiledObj = new JSONObject();
            textFiledObj.put("is_included", isInclude);
            textFiledObj.put("values", new JSONArray(values));
            obj.put(name, textFiledObj);
        }
    }

    /**
     * Represents list field
     *
     */
    static public class ListField {
        private String name;
        private List<String> includeValues;
        private List<String> excludeValues;

        /**
         * Constructs list field
         *
         * @param name filed name
         * @param includeValues include values. Could be null
         * @param excludeValues exclude values Could be null
         */
        public ListField(String name, List<String> includeValues, List<String> excludeValues) {
            this.name = name;
            this.includeValues = includeValues;
            this.excludeValues = excludeValues;
        }

        protected void saveJSONObject(JSONObject obj) throws JSONException {
            JSONObject listFieldsObj = new JSONObject();
            listFieldsObj.putOpt("include", new JSONArray(includeValues));
            listFieldsObj.putOpt("exclude", new JSONArray(excludeValues));
            obj.put(name, listFieldsObj);
        }
    }

    /**
     * Numeric filed
     */
    static public class NumericField {
        private String name;
        private BigDecimal min;
        private BigDecimal max;

        /**
         * Constructs numeric filed
         *
         * @param name filed name
         * @param min minimal value. Could be null
         * @param max maximum value. Could be null
         */
        public NumericField(String name, BigDecimal min, BigDecimal max) {
            this.name = name;
            this.min = min;
            this.max = max;
        }

        protected void saveJSONObject(JSONObject obj) throws JSONException {
            JSONObject longFieldObj = new JSONObject();
            longFieldObj.putOpt("min", min);
            longFieldObj.putOpt("max", max);
            obj.put(name, longFieldObj);
        }
    }


    // ================== Builder ==================

    /**
     * Helper object to build {@link Filter}
     */
    public static class Builder {
        private List<String> includeCategories;
        private List<String> excludeCategories;

        private List<String> l1_categories;
        private List<String> l2_categories;
        private List<String> l3_categories;
        private List<String> l4_categories;

        private BigDecimal minPrice;
        private BigDecimal maxPrice;

        private Integer minDiscount;
        private Integer maxDiscount;

        private List<BooleanField> booleanFields = new LinkedList<>();
        private List<TextField>    textFields = new LinkedList<>();
        private List<ListField>    listFields = new LinkedList<>();
        private List<NumericField> numericFields = new LinkedList<>();

        /**
         * Add include categories
         * @param includeCategories list of categories
         * @return builder (this)
         */
        public Builder withIncludeCategories(List<String> includeCategories) {
            this.includeCategories = includeCategories;
            return this;
        }

        /**
         * Add exclude categories
         * @param excludeCategories list of categories
         * @return builder (this)
         */
        public Builder withExcludeCategories(List<String> excludeCategories) {
            this.excludeCategories = excludeCategories;
            return this;
        }

        /**
         * Add Level 1 categories to filter
         * @param l1_categories category values
         * @return builder (this)
         */
        public Builder withL1_categories(List<String> l1_categories) {
            this.l1_categories = l1_categories;
            return this;
        }

        /**
         * Add Level 2 categories to filter
         * @param l2_categories category values
         * @return builder (this)
         */
        public Builder withL2_categories(List<String> l2_categories) {
            this.l2_categories = l2_categories;
            return this;
        }

        /**
         * Add Level 3 categories to filter
         * @param l3_categories category values
         * @return builder (this)
         */
        public Builder withL3_categories(List<String> l3_categories) {
            this.l3_categories = l3_categories;
            return this;
        }

        /**
         * Add Level 4 categories to filter
         * @param l4_categories category values
         * @return builder (this)
         */
        public Builder withL4_categories(List<String> l4_categories) {
            this.l4_categories = l4_categories;
            return this;
        }

        /**
         * Add minimum price
         * @param minPrice value
         * @return builder (this)
         */
        public Builder withMinPrice(BigDecimal minPrice) {
            this.minPrice = minPrice;
            return this;
        }

        /**
         * Add maximum price
         * @param maxPrice value
         * @return builder (this)
         */
        public Builder withMaxPrice(BigDecimal maxPrice) {
            this.maxPrice = maxPrice;
            return this;
        }

        /**
         * Add discount Range minimum
         * @param minDiscount value in %
         * @return builder (this)
         */
        public Builder withMinDiscount(Integer minDiscount) {
            this.minDiscount = minDiscount;
            return this;
        }

        /**
         * Add discount Range maximum
         * @param maxDiscount value in %
         * @return builder (this)
         */
        public Builder withMaxDiscount(Integer maxDiscount) {
            this.maxDiscount = maxDiscount;
            return this;
        }

        /**
         * Add boolean field filter
         * @param value filed
         * @return builder (this)
         */
        public Builder addBooleanFields(BooleanField value) {
            this.booleanFields.add(value);
            return this;
        }

        /**
         * Add text field filter
         * @param value filed
         * @return builder (this)
         */
        public Builder addTextFields(TextField value) {
            this.textFields.add(value);
            return this;
        }

        /**
         * Add list field filter
         * @param value filed
         * @return builder (this)
         */
        public Builder addListFields(ListField value) {
            this.listFields.add(value);
            return this;
        }

        /**
         * Add numeric field filter
         * @param value filed
         * @return builder (this)
         */
        public Builder addNumericFields(NumericField value) {
            this.numericFields.add(value);
            return this;
        }

        /**
         * Will construct Filter object
         * @return filter object
         */
        public Filter build() {
            return new Filter(this);
        }
    }

}
