package com.jetlore.sdk.ranking;


import android.util.Log;

import com.jetlore.sdk.http.AsyncHttpClient;
import com.jetlore.sdk.common.InvalidUseException;
import com.jetlore.sdk.common.Util;
import com.jetlore.sdk.ranking.callback.GetProductsCallback;

import org.apache.http.Header;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class GetProductsEvent extends RankingGetEvent {

    private static String SHOPPING_CART_MODE = "shopping_cart";

    private String campaignId;
    private Integer tileId;
    private String filter;
    private int limit;
    private List<String> productIds;

    private String clientHash;

    private GetProductsCallback callback;

    private GetProductsEvent(Builder builder) {
        super(builder);
        this.tileId = builder.tileId;
        this.filter = builder.filter;
        this.limit = builder.limit;
        this.campaignId = builder.campaignId;
        this.callback = builder.callback;
        this.clientHash = builder.clientHash;
        this.productIds = builder.productIds;
    }

    protected Map<String, String> getQueryParams() {

        Map<String, String> parametersMap = new HashMap<String, String>();

        if (null != tileId) {
            parametersMap.put("tile", tileId.toString());
        }
        if (!Util.isNullOrEmpty(filter)) {
            parametersMap.put("filter", filter);
        }
        if (!Util.isNullOrEmpty(campaignId)) {
            parametersMap.put("cmpn", campaignId);
        }
        if (null != productIds && !productIds.isEmpty()) {

            StringBuilder productIdsStr = new StringBuilder();
            Iterator<String> productIdIt = productIds.iterator();

            productIdsStr.append(productIdIt.next());
            while(productIdIt.hasNext()) {
                productIdsStr.append(",");
                productIdsStr.append(productIdIt.next());
            }
            parametersMap.put("product_id", productIdsStr.toString());
            parametersMap.put("mode", SHOPPING_CART_MODE);
        }

        parametersMap.put("limit", String.valueOf(limit));
        return parametersMap;
    }

    protected String generateMainGetRequestURL() {
        StringBuilder sb = new StringBuilder();
        sb.append(String.format("%s/offers/offers.json?cid=%s&jl_ctx=mobile",
                apiUrl,
                Util.URLEncode(clientHash)
        ));

        if (null != userId) {
            sb.append(String.format("&id=%s", Util.URLEncode(userId)));
        }

        return sb.toString();
    }

    @Override
    protected void onSuccess(int statusCode, Header[] headers, byte[] responseBody) {
        try {
            JSONObject json = new JSONObject(new String(responseBody, Util.getCharsetFromContentType(headers)));

            List<Product> products = new LinkedList<Product>();

            JSONArray dealsArr = json.optJSONArray("deals");
            if (null == dealsArr) {
                JSONObject deal = json.optJSONObject("deals");
                if (null != deal) {
                    products.add(Product.fromJSONObject(deal));
                }
            } else {
                for (int i = 0; i < dealsArr.length(); ++i) {
                    JSONObject deal = dealsArr.optJSONObject(i);
                    if (null != deal) {
                        products.add(Product.fromJSONObject(deal));
                    }
                }
            }

            callback.onSuccess(products);
        } catch (UnsupportedEncodingException ex) {
            Log.e(TAG, "Failed to parse response from server unsupported content type", ex);
            callback.onFailure("Failed to parse response from server unsupported content type", ex);
        } catch (JSONException ex) {
            Log.e(TAG,"Failed to parse JSON response from server", ex);
            callback.onFailure("Failed to parse JSON response from server", ex);
        }  catch (Exception ex) {
            Log.e(TAG,"Failed to parse JSON unexpected result", ex);
            callback.onFailure("Failed to parse JSON unexpected result", ex);
        }

    }

    @Override
    protected void onFailure(int statusCode, Header[] headers, byte[] responseBody, Throwable error) {
        Log.e(TAG, "Failed to perform request. HTTP status code %n", error);
        callback.onFailure(String.format("Failed to perform request. HTTP status code %n", statusCode), error);
    }

    static class Builder extends RankingEventBuilder<GetProductsEvent> {

        private String campaignId;
        private String filter;
        private Integer limit;
        private Integer tileId;
        private GetProductsCallback callback;
        private String clientHash;
        private List<String> productIds;

        public Builder(String userId, String apiUrl,  String clientHash, AsyncHttpClient httpClient) {
            super(userId, apiUrl, clientHash, httpClient);
            this.clientHash = clientHash;
        }

        public Builder limit(int limit) {
            this.limit = limit;
            return this;
        }

        public Builder filter(String filter) {
            this.filter = filter;
            return this;
        }

        public Builder tile(Integer tileId) {
            this.tileId = tileId;
            return this;
        }

        public Builder campaign(String campaignId) {
            this.campaignId = campaignId;
            return this;
        }

        public Builder productIds(List<String> productIds) {
            this.productIds = new LinkedList<String>(productIds);
            return this;
        }

        public Builder callback(GetProductsCallback callback) {
            this.callback = callback;
            return this;
        }

        @Override
        public GetProductsEvent build() {

            if (null == limit) {
                throw new InvalidUseException("Limit is required parameter");
            }
            if (limit <= 0) {
                throw new InvalidUseException("Limit is can't be negative <= 0");
            }
            if (null == callback) {
                throw new InvalidUseException("Callback is required parameter");
            }

            return new GetProductsEvent(this);
        }

    }
}
