package com.jetlore.sdk.ranking;

import com.jetlore.sdk.ranking.callback.GetProductsCallback;

import org.json.JSONObject;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * Represents product from get Layout product section for {@link ProductSection#getProducts()},
 * {@link PromotionInfo#getFeaturedProducts()}
 * or  {@link Ranker#getProducts(int, RankContext, GetProductsCallback)}
 */
public class Product implements Serializable {

    private String id;
    private String title;
    private String url;
    private String mobileUrl;
    private String imgUrl;
    private String mobileImgUrl;
    private BigDecimal originalPrice;
    private BigDecimal currentPrice;

    private Map<String, String> fields = new HashMap<String, String>();

    /**
     * Gets product Id
     *
     * @return product Identifier
     */
    public String getId() {
        return id;
    }

    /**
     * Gets product url - link to access product
     *
     * @return product url
     */
    public String getUrl() {
        return url;
    }

    /**
     * Gets product mobile url - link that can be used on mobile devises
     *
     * @return product Mobile url
     */
    public String getMobileUrl() {
        return mobileUrl;
    }

    /**
     * Gets product title
     *
     * @return product title
     */
    public String getTitle() {
        return title;
    }

    /**
     * Gets Product image URL. Lower resolution (Desktop resolution)
     *
     * @return - url on image. Null in case if we don't have it
     */
    public String getImgUrl() {
        return imgUrl;
    }

    /**
     * Gets Promotion image URL. High resolution (Mobile resolution)
     *
     * @return - url on image. Null in case if we don't have it
     */
    public String getMobileImgUrl() {
        return mobileImgUrl;
    }

    /**
     * Gets product price before discount
     *
     * @return product price before discount
     */
    public BigDecimal getOriginalPrice() {
        return originalPrice;
    }

    /**
     * Gets product current price
     *
     * @return product current price
     */
    public BigDecimal getCurrentPrice() {
        return currentPrice;
    }

    public Map<String, String> getFields() {
        return fields;
    }

    static Product fromJSONObject(JSONObject obj) {

        Product product = new Product();

        product.id = obj.optString("id", null);
        product.title = obj.optString("title");
        product.url = obj.optString("url", null);
        product.mobileUrl = obj.optString("mobile_url", null);
        product.imgUrl = obj.optString("img", null);
        product.mobileImgUrl = obj.optString("mobile_img", null);

        int intOriginalPrice = obj.optInt("original_price", -1);
        int intCurrentPrice = obj.optInt("current_price", -1);

        product.originalPrice = (-1 == intOriginalPrice) ? null : BigDecimal.valueOf(intOriginalPrice, 2);
        product.currentPrice = (-1 == intCurrentPrice) ? null : BigDecimal.valueOf(intCurrentPrice, 2);

        Iterator<String> keysIterator = obj.keys();
        while(keysIterator.hasNext()) {
            String key = keysIterator.next();
            Object value = obj.opt(key);

            if (value instanceof String) {
                product.fields.put(key, (String) value);
            } else if (value != null) {
                product.fields.put(key, String.valueOf(value));
            } else {
                product.fields.put(key, null);
            }

        }

        return product;
    }
}
