package com.jetlore.sdk.ranking;


import android.content.Context;
import android.os.Bundle;
import android.support.annotation.NonNull;
import android.util.Log;

import com.jetlore.sdk.ranking.callback.SortProductsCallback;
import com.jetlore.sdk.tracking.ActionContext;
import com.jetlore.sdk.common.InvalidUseException;
import com.jetlore.sdk.common.JetloreCommon;
import com.jetlore.sdk.common.Util;
import com.jetlore.sdk.ranking.callback.GetLayoutCallback;
import com.jetlore.sdk.ranking.callback.GetProductsCallback;
import com.jetlore.sdk.ranking.callback.GetPromotionsCallback;
import com.jetlore.sdk.ranking.layout.LayoutConfiguration;

/**
 * Ranker is used to receive ranking events from Jetlore API
 *
 * <p>Ranker needs permission to access the Internet. Add the following to AndroidManifest.xml (inside {@code <manifest>} element):</p>
 * <ul>
 *     <li>{@code <uses-permission android:name="android.permission.INTERNET"/>}</li>
 * </ul>
 *
 * <p>To initialize ActionTracker you need to provide a Client Hash and Jetlore API URL in AndroidManifest.xml:</p>
 * <ul>
 *   <li>Add {@code <string name="jetlore_api_url">https://api.jetlore.com</string>} to strings.xml</li>
 *   <li>Add {@code <meta-data android:name="com.jetlore.sdk.API_URL" android:value="@string/jetlore_api_url"/>} to AndroidManifest.xml (inside {@code <application>} element)</li>
 *   <li>Add {@code <string name="jetlore_client_hash">YOUR_JETLORE_CLIENT_HASH</string>} to strings.xml</li>
 *   <li>Add {@code <meta-data android:name="com.jetlore.sdk.ClientHash" android:value="@string/jetlore_client_hash"/>} to AndroidManifest.xml (inside {@code <application>} element)</li>
 * </ul>
 *
 * or you can call {@link #configure(String, String, String)}  configure(userId, apiUrl, accessToken, String clientHash)
 *
 * <p>You can create an ActionTracker instance in your Application onCreate method like this:</p>
 * <pre><code> {@literal @}Override
 * protected void onCreate(Bundle savedInstanceState) {
 *     super.onCreate(savedInstanceState);
 *     ...
 *     actionTracker = new ActionTracker(this);
 * }</code></pre>
 *
 */
public class Ranker extends JetloreCommon {

    private static final String TAG = "JetloreRanker";

    /** Empty ranking context **/
    public static final RankContext EMPTY_CONTEXT = new RankContext();

    private String _clientHash;

    /**
     * Creates a new Ranker
     *
     * <p>Uses provided {@link android.content.Context} to get Jetlore API access key
     * from AndroidManifest.xml and to get {@link android.content.SharedPreferences} to persist cookies
     *
     * @param context ranking context
     */
    public Ranker(@NonNull Context context) {
        super(context);
    }

    @Override
    protected void readAdditionalParameters(Bundle bundle){

        String clientHash = bundle.getString("com.jetlore.sdk.ClientHash");
        if (Util.isNullOrEmpty(clientHash)) {
            throw new InvalidUseException("Provide your Jetlore API Client Hash token in the AndroidManifest.xml (<meta-data android:name=\"com.jetlore.sdk.ClientHash\" android:value=\"<ClientHash>\"/>)");
        }

        _clientHash = clientHash;
    }

    /**
     * Re-Initialize Ranker with userId, apiUrl and clientHash
     *
     *  Please use {@link #configure(String)} to change or initialize userId
     *  <p>Use {@link #NO_USER} as userId if user id is unknown at the moment.
     *  Make sure to call {@link #configure} again once user id is known.</p>
     *
     *  @param  userId  tracking user id
     *  @param  apiUrl  Jetlore API endpoint URL
     *  @param  clientHash client hash
     *
     *  @throws InvalidUseException
     *                if userId is null or empty
     *                if apiUrl is null or empty
     *                if clientHash is null or empty
     */
    public void configure(@NonNull String userId, @NonNull String apiUrl, @NonNull String clientHash) {

        super.configure(userId, apiUrl);

        Log.d(TAG, String.format("clientHash: %s", clientHash));

        if (Util.isNullOrEmpty(clientHash)) {
            throw new InvalidUseException("clientHash is null or empty");
        }

        _clientHash = clientHash;
    }


    /**
     * An asynchronous method to obtain ranked products for current user from Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param tileId identification of Tile (product image representation)
     * @param filter name of filter that will be used in ranking process
     * @param limit number of products in response
     * @param context {@link RankContext} action context object
     * @param callback {@link GetProductsCallback} callback. Please be aware that callback will be triggered in different thread
     *
     * @throws InvalidUseException
     *                if name of custom parameter {@link RankContext#addCustomParameter(String, String)} intersects with internally used
     *                if limit is 0 or negative
     *                if callback is null
     */
    public void getProducts(Integer tileId, String filter, int limit, RankContext context, GetProductsCallback callback) {
        Log.d(TAG, "getProducts");

        GetProductsEvent.Builder builder = new GetProductsEvent.Builder(getUserId(), _apiUrl, _clientHash, _httpClient);
        fillActionContextInfo(context, builder);

        builder
                .tile(tileId)
                .filter(filter)
                .limit(limit)
                .callback(callback)
                .productIds(context.getProductIds())
                .build()
                .fire();
    }

    /**
     * An asynchronous method to obtain ranked products for current user from Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param filter name of filter that will be used in ranking process
     * @param limit number of products in response
     * @param context {@link RankContext} action context object
     * @param callback {@link GetProductsCallback} callback. Please be aware that callback will be triggered in different thread
     *
     * @throws InvalidUseException
     *                if name of custom parameter {@link RankContext#addCustomParameter(String, String)} intersects with internally used
     *                if limit is 0 or negative
     *                if callback is null
     */
    public void getProducts(String filter, int limit, RankContext context, GetProductsCallback callback) {
        Log.d(TAG, "getProducts");

        GetProductsEvent.Builder builder = new GetProductsEvent.Builder(getUserId(), _apiUrl, _clientHash, _httpClient);
        fillActionContextInfo(context, builder);

        builder
                .filter(filter)
                .limit(limit)
                .callback(callback)
                .productIds(context.getProductIds())
                .build()
                .fire();
    }


    /**
     * An asynchronous method to obtain ranked products for current user from Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param tileId identification of Tile
     * @param limit number of products in response
     * @param context {@link RankContext} action context object
     * @param callback {@link GetProductsCallback} callback. Please be aware that callback will be triggered in different thread
     *
     * @throws InvalidUseException
     *                if name of custom parameter {@link RankContext#addCustomParameter(String, String)} intersects with internally used
     *                if limit is 0 or negative
     *                if callback is null
     */
    public void getProducts(Integer tileId, int limit, RankContext context, GetProductsCallback callback) {
        Log.d(TAG, "getProducts");

        GetProductsEvent.Builder builder = new GetProductsEvent.Builder(getUserId(), _apiUrl, _clientHash, _httpClient);
        fillActionContextInfo(context, builder);

        builder
                .tile(tileId)
                .limit(limit)
                .callback(callback)
                .productIds(context.getProductIds())
                .build()
                .fire();
    }

    /**
     * An asynchronous method to obtain ranked products for current user from Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param limit number of products in response
     * @param context {@link RankContext} action context object
     * @param callback {@link GetProductsCallback} callback. Please be aware that callback will be triggered in different thread
     *
     * @throws InvalidUseException
     *                if name of custom parameter {@link RankContext#addCustomParameter(String, String)} intersects with internally used
     *                if limit is 0 or negative
     *                if callback is null
     */
    public void getProducts(int limit, RankContext context, GetProductsCallback callback) {
        Log.d(TAG, "getProducts");

        GetProductsEvent.Builder builder = new GetProductsEvent.Builder(getUserId(), _apiUrl, _clientHash, _httpClient);
        fillActionContextInfo(context, builder);

        builder
                .limit(limit)
                .callback(callback)
                .productIds(context.getProductIds())
                .build()
                .fire();
    }

    /**
     * An asynchronous method to obtain products sorted by relevance from Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param context {@link SortProductsContext} action context object
     * @param callback {@link SortProductsCallback} callback. Please be aware that callback will be triggered in different thread
     *
     * @throws InvalidUseException
     *                 if callback is null
     *                 if {@link SortProductsContext#filter} or {@link SortProductsContext#productIds} both are nulls
     *                 if {@link com.jetlore.sdk.ranking.SortProductsContext.Pagination} itemsPerPage or page is 0 or negative
     *                 if {@link com.jetlore.sdk.ranking.SortProductsContext.Pagination#itemsPerPage} itemsPerPage is greater then {@link com.jetlore.sdk.ranking.SortProductsContext.Pagination#MAX_ITEMS_PER_PAGE}
     */
    public void sortProductsByRelevance(SortProductsContext context, SortProductsCallback callback) {
        Log.d(TAG, "sortProductsByRelevance");

        SortProductsByRelevanceEvent.Builder builder = new SortProductsByRelevanceEvent.Builder(getUserId(), _apiUrl, _clientHash, _httpClient);
        fillActionContextInfo(context, builder);

        builder
                .filter(context.filter)
                .productIds(context.productIds)
                .pagination(context.pagination)
                .callback(callback)
                .build()
                .fire();

    }

    /**
     * An asynchronous method to obtain {@link Layout}. Layout is combination of personalized products and promotions for current user from Jetlore API.
     *
     * @param layoutName name of layout (from Jetlore Dashboard)
     * @param context {@link ActionContext} action context object
     * @param callback callback. Please be aware that callback will be triggered in different thread
     */
    public void getLayout(String layoutName, ActionContext context, GetLayoutCallback callback) {
        Log.d(TAG, "getLayout");

        GetLayoutEvent.Builder builder = new GetLayoutEvent.Builder(getUserId(), _apiUrl, _clientHash, _httpClient);
        fillActionContextInfo(context, builder);

        builder
                .layoutName(layoutName)
                .callback(callback)
                .build()
                .fire();
    }

    /**
     * An asynchronous method to obtain personalized promotions for current user from Jetlore API.
     *
     * @param collection collection identifier
     * @param tile tile identifier
     * @param limit limit number of promotions in response
     * @param context {@link ActionContext} action context object
     * @param callback callback. Please be aware that callback will be triggered in different thread
     */
    public void getPromos(String collection, String tile, int limit, ActionContext context, final GetPromotionsCallback callback) {
        Log.d(TAG, "getPromos");

        LayoutConfiguration layoutConfiguration = new LayoutConfiguration
                .Builder() // One Section
                    .promotionConfigurationSectionBuilder()
                        .collectionName(collection)
                        .tileName(tile)
                        .limit(limit)
                    .buildAddAndGoUp()
                .build();

        GetLayoutEvent.Builder builder = new GetLayoutEvent.Builder(getUserId(), _apiUrl, _clientHash, _httpClient);
        fillActionContextInfo(context, builder);
        builder
                .layout(layoutConfiguration)
                .callback(new GetLayoutCallback() {
                    public void onSuccess(final Layout layout) {
                        // transform response to GetPromotionsCallback onSuccess
                        // we expect that we have only one promotion section
                        if (null != layout.getSections()
                                && layout.getSections().size() == 1) {
                               layout.getSections().get(0).handle(new SectionVisitor() {
                                   @Override
                                   public void visit(PromotionSection section) {
                                       callback.onSuccess(section.getPromos());
                                   }
                                   @Override
                                   public void visit(ProductSection section) {
                                       String msg = "There is an issue with Jetlore API. Expected promotion section but product section was returned.";
                                       callback.onFailure(msg, new Exception(msg));
                                   }
                               });
                        } else {
                            String msg = "Response from server doesn't contain promotions.";
                            callback.onFailure(msg, new Exception(msg));
                        }
                    }
                    public void onFailure(String msg, Throwable ex) {
                        callback.onFailure(msg, ex);
                    }
                })
                .build()
                .fire();

    }

    private String getUserId() {
        return (_userId != null) ? _userId : NO_USER;
    }
}
