package com.jetlore.sdk.ranking;


import android.util.Log;

import com.jetlore.sdk.common.InvalidUseException;
import com.jetlore.sdk.common.Util;
import com.jetlore.sdk.http.AsyncHttpClient;
import com.jetlore.sdk.ranking.callback.SortProductsCallback;

import org.apache.http.Header;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.UnsupportedEncodingException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class SortProductsByRelevanceEvent extends RankingPostEvent {

    private SortProductsCallback callback;
    private SortProductsContext.Pagination pagination;
    private List<String> productIds;
    private Filter       filter;

    private String BY_FILTER_ENDPOINT  = "products_by_filter.json";
    private String BY_PRODUCT_ENDPOINT = "products_by_ids.json";

    private SortProductsByRelevanceEvent(Builder builder) {
        super(builder);
        this.callback = builder.callback;
        this.pagination = builder.pagination;
        this.productIds = builder.productIds;
        this.filter = builder.filter;
    }

    protected  Map<String, String> getQueryParams(){
        return null;
    }

    @Override
    protected void onSuccess(int statusCode, Header[] headers, byte[] responseBody) {
        try {
            JSONObject json = new JSONObject(new String(responseBody, Util.getCharsetFromContentType(headers)));

            List<String> productIds = new LinkedList<String>();

            JSONArray productIdsArr = json.optJSONArray("product_ids");
            if (null != productIdsArr) {
                for (int i = 0; i < productIdsArr.length(); ++i) {
                    productIds.add(productIdsArr.getString(i));
                }
            }

            callback.onSuccess(productIds, json.optLong("total", -1));
        } catch (UnsupportedEncodingException ex) {
            Log.e(TAG, "Failed to parse response from server unsupported content type", ex);
            callback.onFailure("Failed to parse response from server unsupported content type", ex);
        } catch (JSONException ex) {
            Log.e(TAG,"Failed to parse JSON response from server", ex);
            callback.onFailure("Failed to parse JSON response from server", ex);
        }  catch (Exception ex) {
            Log.e(TAG,"Failed to parse JSON unexpected result", ex);
            callback.onFailure("Failed to parse JSON unexpected result", ex);
        }
    }
    @Override
    protected void onFailure(int statusCode, Header[] headers, byte[] responseBody, Throwable error) {
        Log.e(TAG, "Failed to perform request", error);
        callback.onFailure(String.format("Failed to perform or prepare request. HTTP status code %n", statusCode), error);
    }

    protected String generateMainGetRequestURL(){
        StringBuilder sb = new StringBuilder();
        sb.append(String.format("%s/products/%s?cid=%s&jl_ctx=mobile&jl_page=%s&jl_per_page=%s",
                apiUrl,
                (null != productIds) ? BY_PRODUCT_ENDPOINT : BY_FILTER_ENDPOINT,
                Util.URLEncode(clientHash),
                pagination.pageNo,
                pagination.itemsPerPage
        ));

        if (null != userId) {
            sb.append(String.format("&id=%s", Util.URLEncode(userId)));
        }

        return sb.toString();
    }

    protected byte[] getPostBody() throws Exception {
        JSONObject obj = new JSONObject();

        if (null != filter) {
            obj.put("filter", filter.toJSONObject());
        }
        if (null != productIds) {
            obj.put("product_ids", new JSONArray(productIds));
        }

        return obj.toString().getBytes("UTF-8");
    }


    protected String getBodyContentType(){
        return "application/jsonrequest";
    }


    static class Builder extends RankingPostEvent.RankingEventBuilder<SortProductsByRelevanceEvent> {
        private SortProductsCallback callback;
        private SortProductsContext.Pagination pagination;
        private List<String> productIds;
        private Filter filter;

        public Builder(String userId, String apiUrl,  String clientHash, AsyncHttpClient httpClient) {
            super(userId, apiUrl, clientHash, httpClient);
        }

        public Builder callback(SortProductsCallback callback) {
            this.callback = callback;
            return this;
        }

        public Builder pagination(SortProductsContext.Pagination pagination) {
            this.pagination = pagination;
            return this;
        }

        public Builder filter(Filter filter) {
            this.filter = filter;
            return this;
        }

        public Builder productIds(List<String> productIds) {
            this.productIds = productIds;
            return this;
        }

        @Override
        public SortProductsByRelevanceEvent build() {

            if (null == callback) {
                throw new InvalidUseException("Callback is required parameter");
            }
            if (null != pagination) {
                if (pagination.pageNo <= 0) {
                    throw new InvalidUseException("Pagination ERROR. PageNo parameter can't be negative or zero");
                }
                if (pagination.itemsPerPage <= 0) {
                    throw new InvalidUseException("Pagination ERROR. ItemsPerPage parameter can't be negative or zero");
                }
                if (pagination.itemsPerPage > SortProductsContext.Pagination.MAX_ITEMS_PER_PAGE) {
                    throw new InvalidUseException("Pagination ERROR. Pagination exceeds max items per page (" + SortProductsContext.Pagination.MAX_ITEMS_PER_PAGE + ")");
                }
            } else {
                throw new InvalidUseException("Pagination ERROR. Pagination is required");
            }

            if((null == filter) && (productIds == null)) {
                throw new InvalidUseException("Please specify at least one of parameters filter or productIds");
            }

            return new SortProductsByRelevanceEvent(this);
        }

    }
}
