package com.jetlore.sdk.ranking.layout;

import android.support.annotation.NonNull;

import com.jetlore.sdk.common.InvalidUseException;

import org.json.JSONException;
import org.json.JSONObject;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;


public class PromotionSectionConfiguration extends SectionConfiguration {

    private String tileName;
    private String collectionName;

    List<FeaturedProductsOptions> featuredProductsOptions;

    protected PromotionSectionConfiguration(Builder builder) {
        this(builder.limit,
             builder.customFields,
             builder.tileName,
             builder.collectionName,
             builder.featuredProductsOptions);
    }

    public PromotionSectionConfiguration(@NonNull SectionType type,
                                         int limit,
                                         @NonNull Map<String, String> customFields,
                                         @NonNull String tileName,
                                         @NonNull String collectionName) {
        super(type, limit, customFields);
        if (tileName == null) {
            throw new InvalidUseException("Value of " + tileName + " must be provided");
        }
        if (collectionName == null) {
            throw new InvalidUseException("Value of " + collectionName + " must be provided");
        }
        this.tileName = tileName;
        this.collectionName = collectionName;

        featuredProductsOptions = new LinkedList<FeaturedProductsOptions>();
    }

    public PromotionSectionConfiguration(int limit,
                                         @NonNull Map<String, String> customFields,
                                         @NonNull String tileName,
                                         @NonNull String collectionName,
                                         @NonNull List<FeaturedProductsOptions> featuredProductsOptions) {
        this(SectionType.PROMO, limit, customFields, tileName, collectionName);

        if (featuredProductsOptions != null) {
            for (FeaturedProductsOptions productOptions : featuredProductsOptions) {
                validateFeaturedProductsOption(productOptions);
            }

            this.featuredProductsOptions = featuredProductsOptions;
        }
    }

    private void validateFeaturedProductsOption(FeaturedProductsOptions productOptions) {
        if (productOptions.getPromotionIndex() > limit) {
            throw new InvalidUseException("Specific featured product option has configuration for promotion No(position) " + productOptions.getPromotionIndex() + " that exceeds " + limit + " - maximum number of promotions in section");
        } else if (productOptions.getPromotionIndex() <= 0) {
            throw new InvalidUseException("Specific promotion No(position) must be grater then 0");
        }
    }

    @Override
    JSONObject toJSON(JSONObject obj) throws JSONException{

        obj.put("tile", tileName);
        obj.put("collection", collectionName);

        for (FeaturedProductsOptions productOptions : featuredProductsOptions) {
            obj.put("promo" + productOptions.getPromotionIndex(), productOptions.toJSONObject());
        }

        return obj;
    }

    public void addFeaturedProductsOptions(@NonNull FeaturedProductsOptions option) {
        validateFeaturedProductsOption(option);
        featuredProductsOptions.add(option);
    }


    // ================== Builder ==================
    public static class Builder {
        private int limit;
        private Map<String, String> customFields = new HashMap<String, String>();

        private String tileName;
        private String collectionName;
        List<FeaturedProductsOptions> featuredProductsOptions = new LinkedList<>();

        LayoutConfiguration.Builder parentBuilder;

        protected Builder(LayoutConfiguration.Builder parentBuilder) {
            this.parentBuilder = parentBuilder;
        }

        public Builder tileName(String tileName){
            this.tileName = tileName;
            return this;
        }

        public Builder collectionName(String collectionName){
            this.collectionName = collectionName;
            return this;
        }

        public Builder featuredProductsOptions(List<FeaturedProductsOptions> featuredProductsOptions){
            this.featuredProductsOptions = featuredProductsOptions;
            return this;
        }

        public Builder addFeaturedProductsOption(FeaturedProductsOptions featuredProductsOptions){
            this.featuredProductsOptions.add(featuredProductsOptions);
            return this;
        }

        public Builder limit(int limit) {
            this.limit = limit;
            return this;
        }

        public Builder customField(String name, String value) {
            customFields.put(name, value);
            return this;
        }

        public Builder customFields(Map<String, String> customFields) {
            this.customFields.putAll(customFields);
            return this;
        }

        public FeaturedProductsOptions.Builder featuredProductsOptionsBuilder() {
            return new FeaturedProductsOptions.Builder(this);
        }

        public PromotionSectionConfiguration build() {
            return new PromotionSectionConfiguration(this);
        }

        public LayoutConfiguration.Builder buildAddAndGoUp() {
            parentBuilder.section(build());
            return parentBuilder;
        }
    }
}
