package com.jetlore.sdk.tracking;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Action context object. This object holds list of fields that are common for all method calls
 *
 * <b>feed</b> - is string identification of client feed. Each client can have multiple feeds that corresponds to different retail stores or sections
 * Each feed uses separate recommendation strategies and algorithms
 *
 * <b>labels</b> - list of custom strings that you can use to label tracking events. You can
 * aggregate groups of events with a certain label (e.g., geo region). In Jetlore's dashboard,
 * you can then use provided label values to monitor tracking statistics by custom labels.
 *
 */
public class ActionContext {

    private List<String> labels = new LinkedList<String>();
    private String       feedId;

    private Map<String, String> customParameters = new HashMap<String, String>();

    /**
     * Constructs object without feed identification
     */
    public ActionContext() {}

    /**
     * Constructs object with feed identification
     *
     * @param feedId is string identification of client feed
     */
    public ActionContext(String feedId) {
        this.feedId = feedId;
    }

    /**
     * Add custom label.
     *
     * You can differentiate group of products etc.
     * In our customer dashboard you can use provided label values to monitor custom labeled products statistic
     *
     * @param label label
     * @return same context so you can use this in builder like way  ctx =  .addLabel(...).addLabel(...)
     */
    public ActionContext addLabel(String label){
        labels.add(label);
        return this;
    }

    /**
     * Add custom labels.
     *
     * You can differentiate group of products etc.
     * In our customer dashboard you can use provided label values to monitor custom labeled products statistic
     *
     * @param labels list of labels
     * @return same context so you can use this in builder like way  ctx =  .addLabel(...).addLabel(...)
     */
    public ActionContext addLabels(List<String> labels){
        this.labels.addAll(labels);
        return this;
    }

    /**
     * Adds any custom parameter that you want to be sent with request
     *
     * @param name parameter name
     * @param value parameter value
     * @return same context so you can use this in builder like way
     */
    public ActionContext addCustomParameter(String name, String value){
        customParameters.put(name, value);
        return this;
    }

    /**
     * Method adds any custom parameters that you want to be sent with request
     *
     * @param parameters list of parameters
     * @return same context so you can use this in builder like way
     */
    public ActionContext addCustomParameters(Map<String, String> parameters){
        customParameters.putAll(parameters);
        return this;
    }

    /**
     * Gets list of any custom labels that will be sent with request
     *
     * @return list labels
     */
    public List<String> getLabels() {
        return labels;
    }

    /**
     * Gets feed identification.
     *
     * In case id value is not provided Jelore api will use default one
     *
     * @return feed Id
     */
    public String getFeedId() {
        return feedId;
    }

    /**
     * Gets list of any custom parameters that you want to be sent with request
     *
     * @return list of name value custom request parameters
     */
    public Map<String, String> getCustomParameters() {
        return customParameters;
    }
}
