package com.jetlore.sdk.tracking;

import android.content.Context;
import android.os.Bundle;
import android.support.annotation.NonNull;
import android.util.Log;

import com.jetlore.sdk.common.InvalidUseException;
import com.jetlore.sdk.common.JetloreCommon;
import com.jetlore.sdk.common.Util;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/** ActionTracker is used to send events to Jetlore API
 *
 * <p>ActionTracker needs permission to access the Internet. Add the following to AndroidManifest.xml (inside {@code <manifest>} element):</p>
 * <ul>
 *     <li>{@code <uses-permission android:name="android.permission.INTERNET"/>}</li>
 * </ul>
 *
 * <p>To initialize ActionTracker you need to provide a Jetlore URL and Jetlore API access token in AndroidManifest.xml:</p>
 * <ul>
 *   <li>Add {@code <string name="jetlore_api_url">https://api.jetlore.com</string>} to strings.xml</li>
 *   <li>Add {@code <meta-data android:name="com.jetlore.sdk.API_URL" android:value="@string/jetlore_api_url"/>} to AndroidManifest.xml (inside {@code <application>} element)</li>
 *   <li>Add {@code <string name="jetlore_access_token">YOUR_JETLORE_API_ACCESS_TOKEN</string>} to strings.xml</li>
 *   <li>Add {@code <meta-data android:name="com.jetlore.sdk.AccessToken" android:value="@string/jetlore_access_token"/>} to AndroidManifest.xml (inside {@code <application>} element)</li>
 * </ul>
 *
 * or you can call {@link #configure(String, String, String)}  configure(userId, apiUrl, accessToken, String clientHash)
 *
 * <p>You can create an ActionTracker instance in your activity onCreate method like this:</p>
 * <pre><code> {@literal @}Override
 * protected void onCreate(Bundle savedInstanceState) {
 *     super.onCreate(savedInstanceState);
 *     ...
 *     actionTracker = new ActionTracker(this);
 * }</code></pre>
 *
 * <p>Before using it you need to initialize ActionTracker with a user id: (if you already not initialize it with {@link #configure(String, String, String)})</p>
 * <pre><code>     String userId = ...;
 *     actionTracker.configure(userId);</code></pre>
 * <p>The user ID has to be the same ID you use inside of your email template.</p>
 *
 * <p>Methods  {@link #trackView} and {@link #trackPurchase} are legacy methods. Please use new API {@link #trackView(ProductContext, String, ActionContext)} {@link #trackPurchases(List, ActionContext)} etc</p>
 * <p>{@link #trackView} and {@link #trackPurchase} make asynchronous HTTP calls to Jetlore API, do NOT block and are safe to use from the UI thread.</p>
 **/
public final class ActionTracker extends JetloreCommon {
    /** Constant value to use with {@link #trackView} method if product view doesn't come from email click (deep linking) */
    public final static String NO_CAMPAIGN = "no-campaign";

    /** Empty action context **/
    public final static ActionContext EMPTY_CONTEXT = new ActionContext();

    private static final String TAG = "JetloreActionTracker";

    private String _accessToken;

    /** Creates a new ActionTracker
     *
     * <p>Uses provided {@link android.content.Context} to get Jetlore API access key
     * from AndroidManifest.xml and to get {@link android.content.SharedPreferences} to persist cookies
     *
     * @param context tracking context
     */
    public ActionTracker(@NonNull Context context) {
        super(context);
    }

    @Override
    protected void readAdditionalParameters(Bundle bundle){
        String accessToken = bundle.getString("com.jetlore.sdk.AccessToken");
        if (Util.isNullOrEmpty(accessToken)) {
            throw new InvalidUseException("Provide your Jetlore API access token in the AndroidManifest.xml (<meta-data android:name=\"com.jetlore.sdk.AccessToken\" android:value=\"<access_token>\"/>)");
        }
        _accessToken = accessToken;
    }


    /** Re-Initialize ActionTracker with apiUrl, accessToken and clientHash
     *
     *  Please use {@link #configure(String)} to change or initialize userId
     *  <p>Use {@link #NO_USER} as userId if user id is unknown at the moment.
     *  Make sure to call {@link #configure} again once user id is known.</p>
     *
     *  @param  userId  tracking user id
     *  @param  apiUrl  Jetlore API endpoint URL
     *  @param  accessToken  Client accessToken
     *
     *  @throws InvalidUseException
     *                if userId is null or empty
     *                if apiUrl is null or empty
     *                if accessToken is null or empty
     */
    public void configure(@NonNull String userId, @NonNull String apiUrl, @NonNull String accessToken) {

        super.configure(userId, apiUrl);

        Log.d(TAG, String.format("configure accessToken: %s", accessToken));

        if (Util.isNullOrEmpty(accessToken)) {
            throw new InvalidUseException("accessToken is null or empty");
        }

        _accessToken = accessToken;
    }

    private void checkIsConfigured() {
        if (_userId == null) {
            throw new InvalidUseException("ActionTracker needs to be configured with user id before accessing track methods");
        }
    }

    /** Sends information about product view to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param productId Viewed product ID
     * @param campaignId Jetlore campaign id. Use {@link #NO_CAMPAIGN} if product view doesn't come from email click (deep linking).
     *
     * @deprecated use {@link #trackView(ProductContext, String, ActionContext)} instead.
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if productId is null or empty;
     */
    @Deprecated
    public void trackView(@NonNull String productId,
                          @NonNull String campaignId) {

        Log.d(TAG, String.format("trackView productId: %s, campaignId: %s", productId, campaignId));

        checkIsConfigured();
        new ViewEvent.Builder(_userId, _apiUrl, _accessToken, _httpClient)
                .campaign(campaignId)
                .product(productId)
                .build()
                .fire();
    }


    /** Sends information about product purchase to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param cart Purchased items map, with product id as a key and purchase amount as a value
     *
     * @deprecated use {@link #trackPurchases(List, ActionContext)} instead.
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
    @Deprecated
    public void trackPurchase(@NonNull Map<String, Integer> cart) {
        if (null != cart) {
            List<ProductContext> products = new LinkedList<ProductContext>();
            for (Map.Entry<String, Integer> entry : cart.entrySet()) {
                Long count = (null != entry.getValue()) ? entry.getValue().longValue() : 1L;
                products.add(new ProductContext(entry.getKey(), count));
            }

            trackPurchases(products, new ActionContext());
        }
    }

    /**
     * Sends information about product purchases to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param products list of {@link ProductContext} to sent
     * @param actionContext {@link ActionContext} action context object
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
   public void trackPurchases(List<ProductContext> products,  ActionContext actionContext) {
       Log.d(TAG, String.format("trackPurchases products %n", (null != products) ? products.size() : "null"));

       checkIsConfigured();
       PurchaseProductEvent.Builder builder = new PurchaseProductEvent.Builder(_userId, _apiUrl, _accessToken, _httpClient);

       fillActionContextInfo(actionContext, builder);

       builder.products(products)
               .build()
               .fire();
   }

    /**
     * Sends information about product purchase to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param product  {@link ProductContext} to sent
     * @param actionContext  {@link ActionContext} action context object
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
   public void trackPurchase(ProductContext product,  ActionContext actionContext) {
       List<ProductContext> products = new LinkedList<ProductContext>();
       products.add(product);

       trackPurchases(products, actionContext);
   }

    /**
     * Sends information about product view to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param product  {@link ProductContext} to sent
     * @param actionContext {@link ActionContext} action context object
     */
    public void trackView(ProductContext product,  ActionContext actionContext){
        Log.d(TAG, "trackView product");
        checkIsConfigured();
        ViewEvent.Builder builder = new ViewEvent.Builder(_userId, _apiUrl, _accessToken, _httpClient);

        fillActionContextInfo(actionContext, builder);
        builder
                .product(product)
                .build()
                .fire();
    }

    /**
     * Sends information about product view to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param product  {@link ProductContext} to sent
     * @param campaignId Jetlore campaign id.
     * @param actionContext {@link ActionContext} action context object
     *                                           *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if name of custom parameter {@link ActionContext#addCustomParameter(String, String)} intersects with internally used
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
    public void trackView(ProductContext product, String campaignId, ActionContext actionContext){
        Log.d(TAG, "trackView product with campaign");
        checkIsConfigured();
        ViewEvent.Builder builder = new ViewEvent.Builder(_userId, _apiUrl, _accessToken, _httpClient);

        fillActionContextInfo(actionContext, builder);
        builder
                .campaign(campaignId)
                .product(product)
                .build()
                .fire();

    }

    /** Sends information about Add to Cart to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param products list of {@link ProductContext} to sent
     * @param actionContext {@link ActionContext} action context object
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if name of custom parameter {@link ActionContext#addCustomParameter(String, String)} intersects with internally used
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
    public void trackAddToCartProducts(List<ProductContext> products, ActionContext actionContext) {
        Log.d(TAG, String.format("trackAddToCartProducts product %n", (null != products) ? products.size() : "null"));
        checkIsConfigured();
        AddToCartEvent.Builder builder = new AddToCartEvent.Builder(_userId, _apiUrl, _accessToken, _httpClient);

        fillActionContextInfo(actionContext, builder);
        builder.products(products)
                .build()
                .fire();
    }

    /** Sends information about Add to Cart to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param product {@link ProductContext} to sent
     * @param actionContext {@link ActionContext} action context object
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if name of custom parameter {@link ActionContext#addCustomParameter(String, String)} intersects with internally used
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
    public void trackAddToCartProduct(ProductContext product, ActionContext actionContext) {
        List<ProductContext> products = new LinkedList<ProductContext>();
        products.add(product);

        trackAddToCartProducts(products, actionContext);
    }

    /** Sends information about Remove from Cart to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param products list of {@link ProductContext} to sent
     * @param actionContext {@link ActionContext} action context object
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if name of custom parameter {@link ActionContext#addCustomParameter(String, String)} intersects with internally used
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
    public void trackRemoveFromCartProducts(List<ProductContext> products, ActionContext actionContext) {
        Log.d(TAG, String.format("trackRemoveFromCartProducts product %n", (null != products) ? products.size() : "null"));
        checkIsConfigured();
        RemoveFromCartEvent.Builder builder = new RemoveFromCartEvent.Builder(_userId, _apiUrl, _accessToken, _httpClient);

        fillActionContextInfo(actionContext, builder);

        builder.products(products)
                .build()
                .fire();
    }

    /** Sends information about Remove from Cart to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param product {@link ProductContext} to sent
     * @param actionContext {@link ActionContext} action context object
     *
     * @throws InvalidUseException
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if name of custom parameter {@link ActionContext#addCustomParameter(String, String)} intersects with internally used
     *                if productId is null or empty;
     *                if count id zero or negative;
     */
    public void trackRemoveFromCartProduct(ProductContext product, ActionContext actionContext) {
        List<ProductContext> products = new LinkedList<ProductContext>();
        products.add(product);

        trackRemoveFromCartProducts(products, actionContext);
    }

    /** Sends information about search queries to Jetlore API
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param text search query text
     * @param actionContext {@link ActionContext} action context object
     *
     * @throws InvalidUseException
     *                if name of custom parameter {@link ActionContext#addCustomParameter(String, String)} intersects with internally used
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     */
    public void trackSearchQuery(String text, ActionContext actionContext) {
        Log.d(TAG, "trackSearchQuery");
        checkIsConfigured();

        SearchEvent.Builder builder = new SearchEvent.Builder(_userId, _apiUrl, _accessToken, _httpClient);
        fillActionContextInfo(actionContext, builder);
        builder
                .searchText(text)
                .build()
                .fire();
    }

    /** Special method that is absolutely generic Jetlore API.
     * This method exists only to support backend Jetlore API updates without update of this SDK.
     * We recommend to contact Jetlore Team to get more information about usage (for current API state) or to upgrade your API to Mobile to latest release
     *
     * <p>This method makes an asynchronous HTTP call to Jetlore API, does NOT block and is safe to use from the UI thread</p>
     *
     * @param actionName name of action. Sting that contains identifier of action. To obtain currently supported values please contact Jetlore
     * @param actionContext {@link ActionContext} action context object. Please use {@link ActionContext#addCustomParameter(String, String)} to add parameters
     *
     * @throws InvalidUseException
     *                if name of custom parameter {@link ActionContext#addCustomParameter(String, String)} intersects with internally used
     *                if tracker is not configured (user id is not initialized {@link #configure(String)})
     *                if action is null or empty
     */
    public void trackGenericAction(String actionName, ActionContext actionContext) {
        Log.d(TAG, "trackGenericAction");
        checkIsConfigured();

        GenericEvent.Builder builder = new GenericEvent.Builder(_userId, actionName, _apiUrl, _accessToken, _httpClient);
        fillActionContextInfo(actionContext, builder);

        builder
                .build()
                .fire();
    }
}
