package com.jetlore.sdk.tracking;

import org.json.JSONException;
import org.json.JSONObject;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;

/**
 *  Product context
 *
 *  This object represents product and is used in all track actions as one of input parameters
 */
public class ProductContext {

    private String productId;
    private Long count;
    private BigDecimal pricePerItem;

    private Map<String, String> customParameters = new HashMap<String, String>();

    /**
     * Constructs new ProductContext
     *
     * @param productId - identifier of product
     * @param count - number of products
     * @param pricePerItem - price for single product
     */
    public ProductContext(String productId, Long count, BigDecimal pricePerItem) {
        this.productId = productId;
        this.count = count;
        this.pricePerItem = pricePerItem;
    }

    /**
     * Constructs new ProductContext
     *
     * @param productId - identifier of product
     * @param count - number of products
     */
    public ProductContext(String productId, Long count) {
        this.productId = productId;
        this.count = count;
    }

    /**
     * Constructs new ProductContext
     *
     * @param productId - identifier of product
     */
    public ProductContext(String productId) {
        this.productId = productId;
        this.count = 1l;
    }

    /**
     * Add generic parameter to be transferred with product
     *
     * @param name  name of parameter
     * @param value  value of parameter
     * @return this product context
     */
    public ProductContext addCustomParameter(String name, String value){
        customParameters.put(name, value);
        return this;
    }

    /**
     * Get product ID
     * @return product identifier
     */
    public String getProductId() {
        return productId;
    }

    /**
     * Get Number of products
     * @return  products count
     */
    public Long getCount() {
        return count;
    }

    /**
     * Get price of single product
     * @return price
     */
    public BigDecimal getPricePerItem() {
        return pricePerItem;
    }


    String toJSONString() throws JSONException{

        JSONObject obj = new JSONObject();
        if (null != productId) {
            obj.put("pid", productId);
        }
        if (null != count) {
            obj.put("count", count);
        }
        if (null != pricePerItem) {
            obj.put("price", pricePerItem.toString());
        }

        for (Map.Entry<String, String> parameter : customParameters.entrySet()) {
            obj.put(parameter.getKey(), parameter.getValue());
        }

        return obj.toString().replace("\n","");
    }

}
